/*  message.c -- consistent error, progress and debugging feedback
 *  Copyright (C) 2008, 2009  SEIKO EPSON CORPORATION
 *
 *  License: GPLv2+|iscan
 *  Authors: AVASYS CORPORATION
 *
 *  This file is part of the SANE backend distributed with Image Scan!
 *
 *  Image Scan!'s SANE backend is free software.
 *  You can redistribute it and/or modify it under the terms of the GNU
 *  General Public License as published by the Free Software Foundation;
 *  either version 2 of the License or at your option any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *  You ought to have received a copy of the GNU General Public License
 *  along with this package.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Linking Image Scan!'s SANE backend statically or dynamically with
 *  other modules is making a combined work based on this SANE backend.
 *  Thus, the terms and conditions of the GNU General Public License
 *  cover the whole combination.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  SANE frontends that communicate with Image Scan!'s SANE backend
 *  solely through the SANE Application Programming Interface,
 *  regardless of the license terms of these SANE frontends, and to
 *  copy and distribute the resulting combined work under terms of your
 *  choice, provided that every copy of the combined work is
 *  accompanied by a complete copy of the source code of Image Scan!'s
 *  SANE backend (the version of Image Scan!'s SANE backend used to
 *  produce the combined work), being distributed under the terms of
 *  the GNU General Public License plus this exception.  An independent
 *  module is a module which is not derived from or based on Image
 *  Scan!'s SANE backend.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  independent modules that communicate with Image Scan!'s SANE
 *  backend solely through the "Interpreter" interface, regardless of
 *  the license terms of these independent modules, and to copy and
 *  distribute the resulting combined work under terms of your choice,
 *  provided that every copy of the combined work is accompanied by a
 *  complete copy of the source code of Image Scan!'s SANE backend (the
 *  version of Image Scan!'s SANE backend used to produce the combined
 *  work), being distributed under the terms of the GNU General Public
 *  License plus this exception.  An independent module is a module
 *  which is not derived from or based on Image Scan!'s SANE backend.
 *
 *  Note that people who make modified versions of Image Scan!'s SANE
 *  backend are not obligated to grant special exceptions for their
 *  modified versions; it is their choice whether to do so.  The GNU
 *  General Public License gives permission to release a modified
 *  version without this exception; this exception also makes it
 *  possible to release a modified version which carries forward this
 *  exception.
 */


/*! \file
    \brief  Infra-structure to provide consistent backend feedback.

    \todo  Describe purpose of the three message categories and their
           levels.  Also document usage policy and message formatting
           conventions.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "message.h"


unsigned long msg_level = 0;


#include <ctype.h>
#include <stdlib.h>
#include <strings.h>

/*! Initialises the message infra-structure.

    This function sets level at which the backend should produce
    feedback.  The value is gotten from the \c SANE_DEBUG_EPKOWA
    environment variable.

    The following, case insensitive string literals are supported
    (in increasing level of feedback):

     - \c FATAL
     - \c MAJOR
     - \c MINOR
     - \c INFO
     - \c CALL
     - \c DATA
     - \c CMD
     - \c HEX
     - \c IMG

    \todo  Add support for decimal literal level specification.
 */
void
msg_init (void)
{
  struct level_def
  {
    const char     *key;
    msg_level_type  val;
  };

  const struct level_def def[] =
    {
      {"FATAL", ERR_FATAL},
      {"MAJOR", ERR_MAJOR},
      {"MINOR", ERR_MINOR},

      {"INFO" , LOG_INFO},
      {"CALL" , LOG_CALL},
      {"DATA" , LOG_DATA},

      {"CMD"  , DBG_CMD},
      {"HEX"  , DBG_HEX},
      {"IMG"  , DBG_IMG},

      {NULL}                    /* array terminator */
    };

  const char *level = getenv ("SANE_DEBUG_EPKOWA");
  const struct level_def *p = def;

  msg_level = 0;

  if (!level) return;

  while (p && p->key)
    {
      if (0 == strcasecmp (level, p->key))
        {
          msg_level = p->val;
          log_info ("setting message level to '%s' (%d)",
                    p->key, p->val);
          return;
        }
      ++p;
    }
}


/*! Dumps the contents of a \a buffer in hexadecimal format.
 */
void
msg_dump (const char *fmt, const void *buffer, size_t sz)
{
  const size_t quad_length = 4;
  const size_t quad_count  = 4;
  const size_t line_length = quad_length * quad_count;

  const unsigned char *buf = buffer;

  char ascii[line_length + 1];
  size_t i = 0;

  ascii[line_length] = '\0';

  while (i < sz)
    {
      if (0 == i % line_length)         /* header */
        fprintf (stderr, "%s%08zx: ", fmt, i);

      ascii[i % line_length] = (isprint (buf[i]) ? buf[i] : '.');

      fprintf (stderr, " %02x", buf[i]);
      ++i;
      if (0 == i % quad_length)         /* spacer */
        fprintf (stderr, " ");
      if (0 == i % line_length)         /* trailer */
        fprintf (stderr, " |%s|\n", ascii);
    }

  if (0 != i % line_length)             /* last line */
    {
      do
        {                               /* align trailer */
          ascii[i % line_length] = ' ';
          fprintf (stderr, "   ");
          ++i;
          if (0 == i % quad_length)
            fprintf (stderr, " ");
        }
      while (0 != i % line_length);
      fprintf (stderr, " |%s|\n", ascii);
    }
}
