/*  net-obj.c -- 
 *  Copyright (C) 2008, 2009  SEIKO EPSON CORPORATION
 *
 *  License: GPLv2+|iscan
 *  Authors: AVASYS CORPORATION
 *
 *  This file is part of the SANE backend distributed with Image Scan!
 *
 *  Image Scan!'s SANE backend is free software.
 *  You can redistribute it and/or modify it under the terms of the GNU
 *  General Public License as published by the Free Software Foundation;
 *  either version 2 of the License or at your option any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *  You ought to have received a copy of the GNU General Public License
 *  along with this package.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Linking Image Scan!'s SANE backend statically or dynamically with
 *  other modules is making a combined work based on this SANE backend.
 *  Thus, the terms and conditions of the GNU General Public License
 *  cover the whole combination.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  SANE frontends that communicate with Image Scan!'s SANE backend
 *  solely through the SANE Application Programming Interface,
 *  regardless of the license terms of these SANE frontends, and to
 *  copy and distribute the resulting combined work under terms of your
 *  choice, provided that every copy of the combined work is
 *  accompanied by a complete copy of the source code of Image Scan!'s
 *  SANE backend (the version of Image Scan!'s SANE backend used to
 *  produce the combined work), being distributed under the terms of
 *  the GNU General Public License plus this exception.  An independent
 *  module is a module which is not derived from or based on Image
 *  Scan!'s SANE backend.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  independent modules that communicate with Image Scan!'s SANE
 *  backend solely through the "Interpreter" interface, regardless of
 *  the license terms of these independent modules, and to copy and
 *  distribute the resulting combined work under terms of your choice,
 *  provided that every copy of the combined work is accompanied by a
 *  complete copy of the source code of Image Scan!'s SANE backend (the
 *  version of Image Scan!'s SANE backend used to produce the combined
 *  work), being distributed under the terms of the GNU General Public
 *  License plus this exception.  An independent module is a module
 *  which is not derived from or based on Image Scan!'s SANE backend.
 *
 *  Note that people who make modified versions of Image Scan!'s SANE
 *  backend are not obligated to grant special exceptions for their
 *  modified versions; it is their choice whether to do so.  The GNU
 *  General Public License gives permission to release a modified
 *  version without this exception; this exception also makes it
 *  possible to release a modified version which carries forward this
 *  exception.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "net-obj.h"

#include <errno.h>
#include <limits.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>

#include "defines.h"


const char *net_prog_name = "network";

typedef struct
{
  pid_t pid;
  int   port;
  int   sock;

} net_type;

static net_type *net = NULL;


/*! Start the network plugin in a separate process
 */
static void
net_start (const char *pkglibdir, SANE_Status *status)
{
  int pipe_fd[2];

  SANE_Status s = SANE_STATUS_GOOD;

  if (-1 == pipe (pipe_fd))
    {
      err_fatal ("pipe: %s", strerror (errno));
      if (status) *status = SANE_STATUS_ACCESS_DENIED;
      free (net);
      return;
    }

  net->pid = fork ();
  if (0 == net->pid)
    {
      /*  replace child process with network plugin program
       */
      char prog[PATH_MAX];
      int  n = snprintf (prog, sizeof (prog), "%s%c%s",
                         pkglibdir, FILE_SEP, net_prog_name);

      if (PATH_MAX > n)
        {
          close (pipe_fd[0]);           /* unused read end */
          if (0 <= dup2 (pipe_fd[1], STDOUT_FILENO))
            {
              log_info ("%s[%d]: starting %s", net_prog_name, getpid (),
                        prog);
              if (-1 == execl (prog, prog, NULL))
                {
                  err_fatal ("%s[%d]: %s", net_prog_name, getpid (),
                             strerror (errno));
                }
            }
          else
            {
              err_major ("%s[%d]: %s", net_prog_name, getpid (),
                         strerror (errno));
            }
        }
      else
        {
          err_major ("%s[%d]: %s%c%s: %s", net_prog_name, getpid (),
                     pkglibdir, FILE_SEP, net_prog_name,
                     strerror (ENAMETOOLONG));
        }

      /* notify the parent process that we're done here */
      write (pipe_fd[1], "-1\n", strlen ("-1\n"));
      fsync (pipe_fd[1]);

      close (pipe_fd[1]);
      exit (EXIT_FAILURE);
    }

  if (0 > net->pid)
    {
      err_fatal ("fork: %s", strerror (errno));
      s = SANE_STATUS_CANCELLED;
    }
  else
    {
      /*  Check whether child process has (unexpectedly) exited.  We
          don't want to have zombies in our closet ;-)
       */
      pid_t w = waitpid (net->pid, NULL, WNOHANG);
      if (-1 == w)
        {
          err_minor ("waitpid: %s", strerror (errno));
        }
      if (0 != w)
        {
          log_info ("%s[%d]: exited prematurely", net_prog_name, net->pid);
          s = SANE_STATUS_CANCELLED;
        }
      else
        {
          FILE *fp = fdopen (pipe_fd[0], "rb");
          if (fp)
            {
              if (1 != fscanf (fp, "%d", &(net->port)))
                {
                  err_major ("fscanf: %s", strerror (errno));
                }
              fclose (fp);
            }
          else
            {
              err_fatal ("%s", strerror (errno));
            }
        }
    }
  close (pipe_fd[0]);
  close (pipe_fd[1]);

  if (0 > net->port) s = SANE_STATUS_CANCELLED;

  if (status) *status = s;
}

/*! Connect to the network plugin via a socket
 */
static void
net_connect (SANE_Status *status)
{
  struct sockaddr_in addr;
  struct timeval t;
  int rv = 0;

  log_call ("(%p)", status);

  require (net);

  if (status) *status = SANE_STATUS_GOOD;

  errno = 0;
  net->sock = socket (AF_INET, SOCK_STREAM, 0);
  if (0 > net->sock)
    {
      err_major ("socket: %s", strerror (errno));
      if (status) *status = SANE_STATUS_IO_ERROR;
      return;
    }

  t.tv_sec = 30;
  t.tv_usec = 0;
  errno = 0;
  rv = setsockopt(net->sock, SOL_SOCKET, SO_RCVTIMEO, &t, sizeof (t));
  if (0 > rv)
    {
      err_minor ("socket: %s", strerror (errno));
    }

  errno = 0;
  rv = setsockopt(net->sock, SOL_SOCKET, SO_SNDTIMEO, &t, sizeof (t));
  if (0 > rv)
    {
      err_minor ("socket: %s", strerror (errno));
    }

  memset (&addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons (net->port);
  addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);

  if (0 != connect (net->sock, (struct sockaddr *) &addr, sizeof (addr)))
    {
      err_major ("connect: %s", strerror (errno));
      if (status) *status = SANE_STATUS_IO_ERROR;
      return;
    }
}

void *
net_init (const char *pkglibdir, SANE_Status *status)
{
  SANE_Status s = SANE_STATUS_GOOD;
  int tries = 5;

  log_call ("(%s, %p)", pkglibdir, status);

  if (net)
    {
      err_minor ("been here, done that");
      if (status) *status = s;
      return net;
    }

  if (!pkglibdir) return net;

  net = t_malloc (1, net_type);
  if (!net)
    {
      if (status) *status = SANE_STATUS_NO_MEM;
      return net;
    }

  net->pid  = -1;
  net->port = -1;
  net->sock = -1;

  net_start (pkglibdir, &s);
  if (SANE_STATUS_GOOD != s)
    {
      if (status) *status = s;
      net_exit (net);
      return NULL;
    }

  do
    {
      if (SANE_STATUS_GOOD != s)
        sleep (1);
      net_connect (&s);
    }
  while (0 < --tries && SANE_STATUS_GOOD != s);

  if (SANE_STATUS_GOOD != s)
    {
      if (status) *status = s;
      net_exit (net);
      return NULL;
    }

  if (0 > net->pid || 0 > net->port)
    {
      free (net);
      net = NULL;
    }

  if (status) *status = s;
  return net;
}

void *
net_exit (void *self)
{
  log_call ("(%p)", self);
  require (net == self);

  if (net)
    {
      int status = 0;

      if (0 <= net->sock)
        {
          if (0 != close (net->sock))
            {
              err_minor ("%s", strerror (errno));
            }
        }
      if (1 < net->pid)
        {
          if (0 != kill (net->pid, SIGHUP))
            {
              err_minor ("%s", strerror (errno));
            }
          if (net->pid != waitpid (net->pid, &status, 0))
            {
              err_major ("%s", strerror (errno));
            }

          if (!WIFSIGNALED (status))
            {
              err_major ("%s[%d]: went off the deep end!",
                         net_prog_name, net->pid);
            }
          else
            {
              if (SIGHUP != WTERMSIG (status))
                {
                  err_major ("%s[%d]: %s", net_prog_name, net->pid,
                             strsignal (WTERMSIG (status)));
                }
            }
        }

      free (net);
      net = NULL;
    }

  return net;
}

/*! Obtain the socket connected to the network plugin
 */
int
net_get_sock (void *self)
{
  log_call ("(%p)", self);
  require (net == self);

  if (net) return net->sock;
  return -1;
}
