/* primax_scan -- linux driver for Primax scanners

   $Id: primax_scan.c,v 1.14 2002/11/07 11:45:40 onjo Exp $

   Authors:
   Marco Foglia <Marco.Foglia@switzerland.org>
   Thomas Schano <schano@t-online.de>
   Christian Ordig <chr.ordig@gmx.net>

   Copyright (C) 1999 Marco Foglia, Thomas Schano, Christian Ordig

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/   

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
     
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>            
#include <sched.h>
#include <sys/time.h>
#include <tiffio.h>
#include <getopt.h>
#include <errno.h>

#include "io.h"                        
#include "primax_scan.h"
#include "lp.h"
#include "tables.h"
#include "LM9811.h"
#include <time.h>

int dbg_level = DBG_HIGH;
time_t time1,time2;

#include "libprimax.h"




int main(int argc, char **argv);
static void tiff_open(TIFF **tiffile, scan_image *image);
static void tiff_close(TIFF *tiffile);
void parse_options(scan_image *image, int argc, char **argv);


/*-------------------------------------------------------------*/
/* 
 * opens a tiff file for writing
 */

void tiff_open(TIFF **tiffile, scan_image *image)
{
    *tiffile = TIFFOpen(image->filename, "w");
    if (!tiffile) {
	DBG(DBG_HELPERS, "Error during save: tiff_open did not open file \n");
	return;
    }
    
    TIFFSetField(*tiffile, TIFFTAG_IMAGEWIDTH, image->width);
    TIFFSetField(*tiffile, TIFFTAG_IMAGELENGTH, image->height);
    TIFFSetField(*tiffile, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(*tiffile, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(*tiffile, TIFFTAG_ROWSPERSTRIP,
                 TIFFDefaultStripSize(*tiffile, -1));

    switch( image->colormode)
    {
        case  RGB_MODE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 3);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 8);
            TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, COMPRESSION_LZW &
                         image->lzw);
            break;
        case  GRAY_MODE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, 1);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 1);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 8);
            TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, COMPRESSION_LZW &
                         image->lzw);
            break;
        case  BW_MODE:
        case  BW_MODE_RED:
        case  BW_MODE_GREEN:
        case  BW_MODE_BLUE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, 1);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 1);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 1);
            if ( image->lzw == 1)
                TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, 1);
            else
                TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, 2);
            break;
        case  RED_MODE:
        case  GREEN_MODE:
        case  BLUE_MODE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, 1);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 1);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 8);
            TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, COMPRESSION_LZW &
                         image->lzw);
            break;
    }
    
}


/*-------------------------------------------------------------*/
/* 
 * closes a tiff file 
 */

void tiff_close(TIFF *tiffile)
{
    TIFFClose(tiffile);   
    DBG(DBG_SCAN, "tiff_close: file saved\n");
}


/*-------------------------------------------------------------*/
/* 
 * parse command line input
 */
void parse_options(scan_image *image, int argc, char **argv)
{
    int i, index = 0;
    int c, switch_lamp = 0xFF;
    char *tmp;
    int scanner_model;
    double width = -1, height = -1;
    double temp;


    static struct option options[] = {
	{"help"       ,0, 0, 'h'},
	{"verbose"    ,2, 0, 'v'},
	{"dim"        ,1, 0, 'd'},
	{"res"        ,1, 0, 'r'},
	{"pos"        ,1, 0, 'p'},
	{"version"    ,0, 0, 'V'},
	{"contrast"   ,1, 0, 'c'},
	{"brightness" ,1, 0, 'b'},
	{"gamma"      ,1, 0, 'g'},
	{"file"       ,1, 0, 'f'},
	{"loadcalib"  ,1, 0, 'L'},
	{"writecalib" ,1, 0, 'W'},
	{"lamp"       ,1, 0, 'l'},
	{"RED"        ,0, 0, 'D'},
	{"GREEN"      ,0, 0, 'N'},
	{"BLUE"       ,0, 0, 'E'},
	{"RGB"        ,0, 0, 'R'},
	{"GRAY"       ,0, 0, 'G'},
	{"BW"         ,0, 0, 'B'},
	{"speed"      ,1, 0, 's'},
	{"Scanner"    ,1, 0, 'S'},
	{"Compression",1, 0, 'C'},
	{"Datarate"   ,1, 0, 'K'},
	{0            ,0, 0,   0}
    };
    
    dbg_level = DBG_HELPERS;
    DBG(DBG_HELPERS,"parse_options:\n");
    while (1) {
	 c = getopt_long(argc, argv, "hv:d:Vp:d:r:c:b:g:f:l:DNERGBs:S:C:K:L:W:",
			 options, &index);
	 if (c == -1)
	     break;
	 switch (c){
	 case 'S':
	     RESOLUTION = strtol (optarg, (char **)NULL, 0x00);
	     break;
	 case 'C':
	     if (strtol (optarg, (char **)NULL, 0x00) == 0)
                 image->lzw = 1;
	     break;
	 case 'v':
	     if (!optarg)
		 dbg_level = DBG_HIGH;
	     else
		 dbg_level = strtol (optarg, (char **)NULL, 0x10);
	     DBG(DBG_HELPERS, "  debug level:        0x%02x\n",
                 (unsigned char)dbg_level);
	     break;
	 case 'V':
	     printf("primax_scan version ");
	     printf(VERSION);
	     printf("\n");
	     printf("$Id: primax_scan.c,v 1.14 2002/11/07 11:45:40 onjo Exp $"); 
	     printf("\n");
	     exit(0);
	 case 'p':
	     image->from_left = strtod(optarg, &tmp);
	     image->from_top = strtod(tmp + 1, (char **)NULL);
	     DBG(DBG_HELPERS, "  position:    X= %1.1f x Y= %1.1f inch\n", 
		 image->from_left, image->from_top);
	     break;
	 case 'd':
	     width = strtod(optarg, &tmp);
	     height = strtod(tmp + 1, (char **)NULL);
	     DBG(DBG_HELPERS, "  dimension:   X= %1.1f x Y= %1.1f inch\n", 
		 width, height);
	     break;
	 case 'K':
	     DATA_RATE = strtod(optarg, &tmp);
	     break;
	 case 'r':
	     image->resolution = strtol(optarg, (char **) NULL, 10);
	     DBG(DBG_HELPERS, "  resolution:           %i dpi\n", 
		 image->resolution);
	     break;
	 case 'c':
             temp = strtod(optarg, &tmp);
             for (i = 0; i < 3; i++)
             {
                 image->color[i].contrast = temp;
             }
	     break;
	 case 'b':
	     temp = strtod(optarg, &tmp);
             for (i = 0; i < 3; i++)
             {
                image->color[i].brightness = temp;
             }
	     break;
	 case 'g':
	     temp = strtod(optarg, &tmp);
             for (i = 0; i < 3; i++)
             {
                 image->color[i].gamma = temp;
             }
	     break;
	 case 'f':
	     image->filename = optarg;
	     DBG(DBG_HELPERS, "  filename:     %s\n",image->filename); 
	     break;
	 case 'L':
	     image->load_calibration_file = optarg;
	     break;
	 case 'W':
	     image->save_calibration_file = optarg;
	     break;
	 case 'D':
	     image->colormode = (image->colormode & 0xF0F1) | 0x0011;
	     break;
	 case 'E':
	     image->colormode = (image->colormode & 0xF0F4) | 0x0014;
	     break;
	 case 'N':
	     image->colormode = (image->colormode & 0xF0F2) | 0x0012;
	     break;
	 case 'R':
	     image->colormode = (image->colormode & 0xFF0F) | 0x0007;
	     break;
	 case 'G':
	     image->colormode = (image->colormode & 0xFF0F) | 0x0107;
	     break;
	 case 'B':
             if ( image->colormode & 0x0010 )
                 image->colormode = (image->colormode & 0xF0FF) | 0x1000;
             else
                 image->colormode = BW_MODE; 
	     break;
	 case 's':
	     image->speed = (int) strtol (optarg, (char **)NULL, 0) & 0xf;
	     break;
	 case 'l':
	     if (strcmp((const char *)optarg, "off"))
		 switch_lamp = 0x20;
	     else
		 switch_lamp = 0x00;
	     DBG(DBG_HELPERS, "  lamp:                 %s\n",optarg); 
	     break;
	 case 'h':
	     printf("\n\nUsage:\n");
	     printf("-h         ,   --help            shows this message\n");
	     printf("-v [LEVEL] ,   --verbose=[LEVEL] increase debugging "
                    "output\n");
	     printf("-s SPEED   ,   --speed=SPEED     SPEED of the carriage "
                    "{0,1,2...15}\n"
                    "                                 (Default auto-ajusted "
                    "with datarate)\n");
	     printf("-K RATE    ,   --Datarate=RATE   Max. Datarate of "     
                    "EPP-port (%.2f kB/s)\n",DATA_RATE);
	     printf("-S RES     ,   --Scanner=RES     max. optical RESolution "
                    "of the Scanner\n"
                    "                                 {300,600} dpi\n");
	     printf("-V         ,   --version         display version\n");
	     printf("-p LxT     ,   --pos=LxT         position form Left "
                    "x Top in inch\n");
	     printf("-d WxH     ,   --dim=WxH         Width x Heigth of "
                    "scan area in\n"
                    "                                 inch (max. %2.2f"
                    " x %2.2f)\n", MAX_WIDTH, MAX_HEIGHT);
	     printf("-c C       ,   --contrast=C      Contrast "
                    "{(-2000)..(1000)}\n");
	     printf("-b B       ,   --brightness=B    Brightness "
                    "{(-200)..(200)}\n");
	     printf("-g G       ,   --gamma=G         Gamma {(0.2)..(5)}\n");
	     printf("-r RES     ,   --res=RES         RESolution in dpi (max. "
                    "%d) see option -S\n", RESOLUTION);
	     printf("-f NAME    ,   --file=NAME       NAME for tiff-file\n");
	     printf("-W NAME    ,   --writecalib=NAME writes calibration to file NAME\n");
	     printf("-L NAME    ,   --loadcalib=NAME  loads calibration from file NAME\n");
	     printf("-C val     ,   --Compression=val val=0 switches "
                    "compression off\n");
	     printf("-l on/off  ,   --lamp=on/off     turn lamp on/off\n");
	
             printf("\n");
             printf("-D         ,   --RED             only RED-channel\n");
             printf("-E         ,   --BLUE            only BLUE-channel\n");
             printf("-N         ,   --GREEN           only GREEN-channel\n");
             printf("-R         ,   --RGB             24-Bit RGB-mode "
                    "(default)\n");
             printf("-G         ,   --GRAY            8-Bit GRAY-mode\n");
             printf("-B         ,   --BW              1-Bit Black-White-"
                    "mode (possible with\n"
                    "                                 all other modes)\n");
	     printf("\n\n\n");
	     exit(0);
	 }
    }
    if ((width != -1) || (height != -1))
        {
            image->width = (int)(image->resolution * width);
            image->height = (int)(image->resolution * height);
        }
    
    DBG(DBG_HELPERS, "  dimension:        %3ix%3i pixel\n", 
	image->width, image->height);

    /* switch lamp and try to detect scanner */
    if (switch_lamp != 0xFF) {
	DBG(DBG_HELPERS, "parse_options:       switching lamp\n");
	if (switch_lamp == 0x00) 
	    scanner_model = lamp_off();
	else
	    scanner_model = lamp_on();
	if ( scanner_model >= 0) {
            switch ( scanner_model & 0x0f ) {
                case 0x02:
                    printf("\nLooks like a Primax Colorado Direct\n");
                    break;
                case 0x05:
                    printf("\nLooks like a Primax Colorado Direct 600\n");
                    break;
                case 0x44:
                    printf("\nLooks like a Primax Colorado D600_OEM\n");
                    break;
                default:
                    printf("\nCan not identify this scanner. code: %02x\n", scanner_model);
	    }
	}
	else {
	    printf("\n:-(\nScanner did not switch to EPP mode");
	    printf(".....not compatible?\n");
	    exit(1);
	}
	exit(0);
    }
}


void scan_tiff(scan_image *image){
    
    TIFF *tiffile;
    int y = 0;
    int size = 0;
    unsigned char *data;
    struct scan_session sc;
     
    tiff_open(&tiffile, image);
    prepare_scanner(image, SCAN);
    
    prepare_scan(&sc, image);

    do {
	data = get_scan_data(&sc, &size);
	if (size < 0) break;
	TIFFWriteScanline(tiffile, data, y , 0);
	y++;
    } while (size >= 0);
    
    finish_scan(&sc);
    finish_scanner();
    tiff_close(tiffile);  
}

/*-------------------------------------------------------------*/

int init_LM9811(void);

int main(int argc, char **argv)
{
    scan_image *image;
    /*    char i,j; */
    int calibration_success;


    init_image( &image);
    parse_options(image, argc, argv);
/*      calc_x_table(image); */
/*      calc_y_table(image); */
/*      if(image->speed_flag) */
    calc_speed(image);
    
//  /dev/lpx must be allocated!
    if ( epp_on() != EPP_ON_OK )
	printf("cannot access parallel port\n");
    
    wakeup();
/*           printf("\n"); */
    power_sequence(CARRIGE_RETURN);
    wait_for_carriage();

    init_LM9811();

    calibration_success = 0;
    if (image->load_calibration_file)
	if ((calibration_success = load_calibration(image))==0) {
	    
	    if (errno)
		perror(image->load_calibration_file);
	    else
		fprintf(stderr, "%s: file read error\n", image->load_calibration_file);
	}
    if (!calibration_success) {
	printf("calibration\n");
	calibrate_LM9811(image);
    }
    if (image->save_calibration_file)
	if (!save_calibration(image))
	    perror(image->save_calibration_file);
   
    
   
    power_sequence(CARRIGE_RETURN);
    power_sequence(S_4000);
    /*          for (j = 0; j < 0x10; j++) */
    /*          { */
    /*              epp_read_one(&i,j); */
    /*              printf("%x,%2x|",j,i&0xff ); */
    /*          } */
    /*          printf("\n"); */
    time(&time1);
    scan_tiff(image);
    time(&time2);
       
    /*          printf("time: %f sec\n",difftime(time2,time1));  */
    power_sequence(CARRIGE_RETURN);
    /*          for (j = 0; j < 0x10; j++) */
    /*          { */
    /*              epp_read_one(&i,j); */
    /*              printf("%x,%2x|",j,i&0xff ); */
    /*          } */
    /*          printf("\n"); */
    power_sequence(GO_OFFLINE);
    epp_off();
    free_image(image);
    return 0;
}  



  
/* This stuff is for emacs
 * Local variables:
 * c-basic-offset: 4
 * End:       
 */
